# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Trax decorators and layers for asserts on tensor shapes."""

import functools
import inspect
import string
from absl import logging
from trax.layers import base
from trax.layers import combinators


def assert_shape(specification):
  """Decorator for checking the input and output shapes of Layer.

  Decorator can be applied on trax.base.Layer class, or a function returning
  a trax.base.Layer class. It uses notation similar to einsum (Einstein
  summation convention), achieving concise and simple representation of tensors.
  For example 'ij,jh->ih' is a valid representation of a function taking two
  2D matrices as input, and returning a single output, also a 2D matrix.

  It improves readability and puts puts three levels of asserts on the function:
  first level is the number of input tensors and output tensors; second level is
  the rank of each tensor; third level is the size of each dimension of each
  tensor. The decorator inserts those asserts right before and right after
  'forward' call.

  First level, assert on number of inputs and outputs. In the representation
  input tensors are separated from output tensors by an arrow '->'. For layers
  taking multiple input tensors or returning multiple output tensors, those
  tensors will be separated by a comma ','.
  For example, specification 'bsd,df->bsf' asserts that there will be two
  input tensors, with shapes represented by 'bsd' and 'df' respectively; and
  a single output tensor with shape represented by 'bsf'.

  Second level, asserts on possible rank of each tensor. Most commonly,
  each letter represents a single dimension. For example,the tensor with shapes
  represented by 'bsd' has rank three; with 'df' it has rank two. The special
  case is an ellipsis ('...'), which expand to arbitrary number of dimensions,
  including zero. For example, the tensor with specification '...sf' has at
  least two dimensions. Each tensor may have in its representation one ellipsis.

  Third level, asserts the size of each dimension. If two dimensions in any
  of input or output tensors have the same letter in the representation then
  they must have the same size. For example, with a tensor A represented by 'df'
  and a tensor B represented by 'bsf', the size of the second dimension of A
  must equal the size of the third dimension of B. Another example: with a
  tensor C represented by '...dv' and a tensor D represented by 'd', the size of
  the first and only dimension of D must be equal to the size of the second to
  last dimension of tensor C.

  If two distinct tensors have an ellipsis in their representation then all of
  dimensions covered by those ellipses must match. For example, with a tensor E
  represented by '...d' and tensor F represented by '...x' then E and F must
  have the same rank, and the sizes of all but the last dimensions must match.

  Examples:
  # In Dense layer there is a single input and single output; the last dimension
  # may change in size, while the sizes of all previous dimensions, marked by
  # an ellipsis, will stay the same.
  @assert_shape('...a->...b')
  class Dense(base.Layer):
    (...)

  # DotProductCausalAttention takes three tensors as input: Queries, Keys, and
  # Values, and outputs a single tensor. Sizes of the first two dimensions in
  # all those tensors must match, while the last dimension must match only
  # between Queries and Keys, and separately between Values and output tensor.
  @assert_shape('blk,blk,bld->bld')
  class DotProductCausalAttention(base.Layer):
    (...)

  # assert_shape can also be placed before the function returning base.Layer.
  @assert_shape('...d->...')
  def ReduceSum():
    return Fn('ReduceSum', lambda x: jnp.sum(x, axis=-1, keepdims=False))

  Args:
    specification: A text specification for the input/output tensors.

  Returns:
    The decorator changing the class or function.
  """
  caller = inspect.getframeinfo(inspect.stack()[1][0])
  message = f'Defined at {caller.filename}:{caller.lineno}'

  def wrap_cls(cls):
    forward = getattr(cls, 'forward')
    init = getattr(cls, '__init__')

    before_spec, after_spec = specification.split('->')

    @functools.wraps(init)
    def init_wrapper(self, *args, **kwargs):
      before_assert = AssertShape(before_spec,
                                  message=message + ' function input')
      after_assert = AssertShape(after_spec,
                                 message=message + ' function output')
      after_assert._create_link(before_assert)  # pylint: disable=protected-access
      out = init(self, *args, **kwargs)
      self._before_assert_fun = before_assert  # pylint: disable=protected-access
      self._after_assert_fun = after_assert  # pylint: disable=protected-access
      return out

    @functools.wraps(forward)
    def forward_wrapper(self, x, *args, **kwargs):
      x = self._before_assert_fun.forward(x)  # pylint: disable=protected-access
      y = forward(self, x, *args, **kwargs)
      y = self._after_assert_fun.forward(y)  # pylint: disable=protected-access
      return y

    setattr(cls, 'forward', forward_wrapper)
    setattr(cls, '__init__', init_wrapper)
    return cls

  # TODO(jaszczur): replace this with forward/init override.
  def wrap_fun(fun):
    @functools.wraps(fun)
    def fun_wrapper(*args, **kwargs):
      layer = fun(*args, **kwargs)
      return AssertFunction(specification, layer, message)
    return fun_wrapper

  def wrap_fun_or_cls(fun_or_cls):
    return (wrap_cls(fun_or_cls) if inspect.isclass(fun_or_cls) else
            wrap_fun(fun_or_cls))

  return wrap_fun_or_cls


def AssertFunction(specification, layer, message=None):  # pylint: disable=invalid-name
  """AssertFunction asserts shapes on the input/output tensors of a layer.

  It passes all inputs to the layer, and returns all outputs of the layer
  unchanged.

  Args:
    specification: A specification. See assert_shape decorator for a full
        documentation.
    layer: A base.Layer to wrap around.
    message: An optional message to print if an assert fails. By default it will
        print the filename and the line number where AssertFunction was called.

  Returns:
    The given layer wrapped in asserts on its inputs and outputs.
  """
  if message is None:
    caller = inspect.getframeinfo(inspect.stack()[1][0])
    message = f'Defined at {caller.filename}:{caller.lineno}'
  before_spec, after_spec = specification.split('->')
  before_assert = AssertShape(before_spec, message=message + ' function input')
  after_assert = AssertShape(after_spec, message=message + ' function output')
  after_assert._create_link(before_assert)  # pylint: disable=protected-access
  return combinators.Serial(
      before_assert, layer, after_assert)


class AssertShape(base.Layer):
  """Layer which put asserts on shapes of tensors, and returns them unchanged.

  It borrows the notation from assert_shape decorator, except it doesn't have
  the arrow '->' special character, as the input tensors are the same as output.
  """

  def __init__(self, spec, message=None, visible_layer=False):
    """Creates AssertShape layer.

    Args:
      spec: Specification for input tensors. See assert_shape decorator for the
          full documentation.
      message: An optional message to include when assert fails. By default it
          includes the filename and line number where this function was called.
      visible_layer: If true, print this layer inside the model (default: False)
    """
    name = 'AssertShape' if visible_layer else ''
    super().__init__(name=name)
    spec = spec.replace('...', '.')
    for letter in spec:
      assert letter in string.ascii_letters + string.digits + '.' + ','
    self._specs = spec.split(',')
    self._n_in = self._n_out = len(self._specs)

    self._defined_shapes = {str(i): i for i in range(10)}
    self._linked = False

    if message is None:
      caller = inspect.getframeinfo(inspect.stack()[1][0])
      self._message = f'Defined at {caller.filename}:{caller.lineno}'
    else:
      self._message = message

  def forward(self, xs):
    if not self._linked:
      for k in list(self._defined_shapes.keys()):
        if not k.isdigit():
          del self._defined_shapes[k]

    if not isinstance(xs, (list, tuple)):
      xs = [xs]

    # Try-except below checks if something is wrong with shapes. It can happen
    # e.g. when using trax2keras. If this is the case we cannot check if shapes
    # are correct or not
    try:
      for x in xs:
        for i in range(len(x.shape)):
          if x.shape[i] != x.shape[i]:
            raise TypeError()
    except TypeError:
      message = ('AssertShape cannot check shapes. This often happens when'
                 ' using trax2keras. Shape asserts are skipped.')
      print(message)
      logging.warning(message)
      if len(xs) == 1:
        return xs[0]
      else:
        return xs

    # helper functions
    def assert_true(cond):
      if not cond:
        shapes = [x.shape for x in xs]
        defined_shapes_dict_without_digits = {
            k: v for k, v in self._defined_shapes.items() if not k.isdigit()}
        raise ValueError(
            f'AssertShape Error. Expected {self._specs}, got {shapes} with dict'
            f' {defined_shapes_dict_without_digits}. {self._message}')

    def assert_equal(a, b):
      assert_true(a == b)
      return a

    def check_shape(shape, spec):
      assert_equal(len(shape), len(spec))
      for shape_dim, letter in zip(shape, spec):
        if letter in self._defined_shapes:
          self._defined_shapes[letter] = assert_equal(
              self._defined_shapes[letter], shape_dim)
        else:
          self._defined_shapes[letter] = shape_dim

    def check_ellipsys(shape):
      if '.' not in self._defined_shapes:
        self._defined_shapes['.'] = shape
      else:
        assert_equal(len(shape), len(self._defined_shapes['.']))
        for s1, s2 in zip(shape, self._defined_shapes['.']):
          assert_equal(s1, s2)

    # actual asserts
    assert_equal(len(xs), len(self._specs))

    for x, spec in zip(xs, self._specs):
      if '.' in spec:
        assert_true(len(x.shape) >= (len(spec) - 1))

        before, after = spec.split('.')
        check_shape(x.shape[:len(before)], before)
        if after:
          check_shape(x.shape[-len(after):], after)
          check_ellipsys(x.shape[len(before):-len(after)])
        else:
          # if len(after) == 0 then -len(after) in indices evaluates badly.
          check_ellipsys(x.shape[len(before):])
      else:
        check_shape(x.shape, spec)

    if len(xs) == 1:
      return xs[0]
    else:
      return xs

  def _create_link(self, other):
    """Internal. Used to create a shared dictionary."""
    # This works well for assert_shape and AssertFunction; but it can break
    # easily if the order of calls to forward() is not known in advance.
    self._linked = True
    self._defined_shapes = other._defined_shapes  # pylint: disable=protected-access



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Loop callbacks.

Callbacks can be used to customize the behavior of `supervised.training.Loop`
to accomodate a variety of use-cases.

Examples include:
  - custom evaluation schemes
  - logging metrics to external servers
  - sending model checkpoints to external servers
  - updating the target network in RL algorithms and other non-stationary
    problems
"""

import collections
import os

import gin
import numpy as np

from trax import jaxboard
from trax import layers as tl
from trax import shapes
from trax.rl import serialization_utils
from trax.supervised import decoding


class TrainingStepCallback:
  """Callback triggered before and after a training step."""

  def __init__(self, loop):
    """Initializes the callback with a `supervised.training.Loop` instance."""
    self._loop = loop

  def call_at(self, step):
    """Returns whether the callback should be called at a given step."""
    raise NotImplementedError

  def on_step_begin(self, step):
    """Called by Loop before training steps, when call_at returned True."""
    raise NotImplementedError

  def on_step_end(self, step):
    """Called by Loop after training steps, when call_at returned True."""
    raise NotImplementedError


@gin.configurable
class SerializedModelEvaluation(TrainingStepCallback):
  """Evaluates serialized sequence prediction models.

  Example: time series prediction. We can serialize a time series into
  a sequence of discrete tokens and model this sequence using an autoregressive
  sequence model, such as Transformer - see
  `trax.rl.serialization_utils.SerializedModel`. Then we can use this callback
  to evaluate long-horizon predictions of such a model.
  """

  def __init__(
      self,
      loop,
      model=None,
      eval_at=1000,
      eval_task=None,
      context_lengths=(1,),
      horizon_lengths=(1,),
      n_steps=1,
      accelerate_model=True,
  ):
    """Initializes SerializedModelEvaluation.

    Args:
      loop: Instance of `trax.supervised.training.Loop` or `None`. Can be set to
        `None` for testing - in such a case, `model` and `eval_task` must be
        provided.
      model: Instance of `trax.rl.serialization_utils.SerializedModel`. Not
        required if `loop` is provided.
      eval_at: When to evaluate. Either int (every how many steps to evaluate),
        or a list of ints (step numbers), or a function int -> bool (step
        predicate).
      eval_task: Instance of `trax.supervised.training.EvalTask` with the
        evaluation data, or None. If not provided, the task will be taken from
        `loop`.
      context_lengths: List of lengths of the context sequence fed into the
        model before starting prediction.
      horizon_lengths: List of lengths of the predicted sequence.
      n_steps: Number of batches to run evaluation for.
      accelerate_model (bool): Whether to wrap the model in `tl.Accelerate`.
    """
    super().__init__(loop)

    if model is None:
      model = loop.model

    observation_serializer = model.observation_serializer
    action_serializer = model.action_serializer

    predict_model = model.make_predict_model()
    if accelerate_model:
      predict_model = tl.Accelerate(predict_model)
    self._predict_model = predict_model
    self._obs_serializer = observation_serializer
    self._act_serializer = action_serializer

    if isinstance(eval_at, int):
      self._eval_at = lambda step: step % eval_at == 1
    elif hasattr(eval_at, '__in__'):
      self._eval_at = lambda step: step in eval_at
    elif callable(eval_at):
      self._eval_at = eval_at
    else:
      raise TypeError(f'Unsupported type for eval_at: {type(eval_at)}.')

    if eval_task is None:
      if len(loop.eval_tasks) != 1:
        raise ValueError(
            'If eval_task is not provided, the number of eval_tasks registered '
            'in Loop must be exactly 1.'
        )
      eval_task = loop.eval_tasks[0]
    self._eval_task = eval_task

    self._context_lengths = list(sorted(context_lengths))
    self._horizon_lengths = list(sorted(horizon_lengths))
    self._n_steps = n_steps

    self._batch_size = eval_task.sample_batch[0].shape[0]
    (_, self._init_state) = predict_model.init(
        shapes.ShapeDtype((self._batch_size, 1), dtype=np.int32)
    )

  @property
  def predict_model(self):
    return self._predict_model

  def call_at(self, step):
    return self._eval_at(step)

  def on_step_begin(self, step):
    pass

  def on_step_end(self, step):
    summary_writer = jaxboard.SummaryWriter(
        os.path.join(self._loop.output_dir, 'srl_eval')
    )
    try:
      weights = self._loop.eval_model.seq_model_weights
      metrics = self.evaluate(weights)
      self._loop.log_summary(metrics, summary_writer, '', 'srl_eval')
    finally:
      summary_writer.close()

  def evaluate(self, weights):
    """Evaluates the model and returns the metrics."""
    self._predict_model.weights = weights

    metrics = collections.defaultdict(list)
    for _ in range(self._n_steps):
      batch = self._eval_task.next_batch()
      step_metrics = self._evaluate_batch(batch)
      for (key, value) in step_metrics.items():
        metrics[key].append(value)

    metrics = {k: np.array(v) for (k, v) in metrics.items()}

    def metric_name(context, horizon):
      return f'pred_error/context_{context}/horizon_{horizon}'

    return {
        metric_name(context, horizon):
            np.sum(errors) / (np.sum(errors != 0) + 1e-6)
        for ((context, horizon), errors) in metrics.items()
    }

  def _evaluate_batch(self, batch):
    """Performs evaluation on a single batch."""
    (obs, act, _, mask) = batch
    obs_repr = serialization_utils.Serialize(self._obs_serializer)(obs)
    act_repr = serialization_utils.Serialize(self._act_serializer)(act)

    errors = {}
    last_context = 0
    last_state = self._init_state
    last_start_id = 0
    for context in self._context_lengths:
      self._predict_model.state = last_state
      start_id = last_start_id

      if context > last_context:
        context_seq = serialization_utils.Interleave()((
            obs_repr[:, last_context:context], act_repr[:, last_context:context]
        ))
        consume_sequence(self._predict_model, start_id, context_seq[:, :-1])
        last_start_id = start_id = context_seq[:, -1:]
        last_state = self._predict_model.state
        last_context = context

      for timestep in range(max(self._horizon_lengths)):
        pred_repr = decoding.autoregressive_sample(
            self._predict_model,
            start_id=start_id,
            eos_id=-1,
            batch_size=self._batch_size,
            max_length=self._obs_serializer.representation_length,
            accelerate=False,
        )
        horizon = timestep + 1
        if horizon in self._horizon_lengths:
          pred = self._obs_serializer.deserialize(pred_repr)
          error = self._calculate_error(pred, obs[:, context + timestep])
          errors[context, horizon] = error * mask[:, context + timestep]

        start_id = pred_repr[:, -1:]
        consume_sequence(
            self._predict_model, start_id, act_repr[:, context + timestep, :-1]
        )
        start_id = act_repr[:, context + timestep, -1:]

    return errors

  def _calculate_error(self, prediction, ground_truth):
    return (prediction - ground_truth) ** 2


def consume_sequence(model, start_id, sequence):
  decoding.autoregressive_sample(
      model,
      start_id=start_id,
      eos_id=-1,
      inputs=sequence,
      batch_size=sequence.shape[0],
      max_length=1,
      accelerate=False,
  )



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Core layer types and key functions used by various layers."""

from absl import logging
import numpy as np
import tensorflow as tf

from trax import fastmath
from trax.fastmath import numpy as jnp
from trax.layers import base
from trax.layers import initializers as init
from trax.layers.assert_shape import assert_shape
from trax.layers.base import Fn


# The output tensor has the same shape as the input tensor, except for the size
# of the last dimension.
@assert_shape('...a->...b')
class Dense(base.Layer):
  """A dense (a.k.a. fully-connected, affine) layer.

  Dense layers are the prototypical example of a trainable layer, i.e., a layer
  with trainable weights. Each node in a dense layer computes a weighted sum of
  all node values from the preceding layer and adds to that sum a node-specific
  bias term. The full layer computation is expressed compactly in linear
  algebra as an affine map `y = Wx + b`, where `W` is a matrix and `y`, `x`,
  and `b` are vectors. The layer is trained, or "learns", by updating the
  values in `W` and `b`.

  Less commonly, a dense layer can omit the bias term and be a pure linear map:
  `y = Wx`.
  """

  def __init__(self,
               n_units,
               kernel_initializer=init.GlorotUniformInitializer(),
               bias_initializer=init.RandomNormalInitializer(1e-6),
               use_bias=True,
               use_bfloat16=False):
    """Returns a dense (fully connected) layer of width `n_units`.

    A dense layer maps collections of `R^m` vectors to `R^n`, where `n`
    (`= n_units`) is fixed at layer creation time, and `m` is set at layer
    initialization time.

    Args:
      n_units: Number of nodes in the layer, also known as the width of the
          layer.
      kernel_initializer: Function that creates a matrix of (random) initial
          connection weights `W` for the layer.
      bias_initializer: Function that creates a vector of (random) initial
          bias weights `b` for the layer.
      use_bias: If `True`, compute an affine map `y = Wx + b`; else compute
          a linear map `y = Wx`.
      use_bfloat16: If `True`, use bfloat16 weights instead of the default
        float32; this can save memory but may (rarely) lead to numerical issues.
    """
    super().__init__(name=f'Dense_{n_units}')
    self._n_units = n_units
    self._kernel_initializer = kernel_initializer
    self._bias_initializer = bias_initializer
    self._use_bias = use_bias
    self._use_bfloat16 = use_bfloat16

  def forward(self, x):
    """Executes this layer as part of a forward pass through the model.

    Args:
      x: Tensor of same shape and dtype as the input signature used to
          initialize this layer.

    Returns:
      Tensor of same shape and dtype as the input, except the final dimension
      is the layer's `n_units` value.
    """
    if self._use_bias:
      if not isinstance(self.weights, (tuple, list)):
        raise ValueError(f'Weights should be a (w, b) tuple or list; '
                         f'instead got: {self.weights}')
      w, b = self.weights
      return jnp.dot(x, w) + b  # Affine map.
    else:
      w = self.weights
      return jnp.dot(x, w)  # Linear map.

  def init_weights_and_state(self, input_signature):
    """Randomly initializes this layer's weights.

    Weights are a `(w, b)` tuple for layers created with `use_bias=True` (the
    default case), or a `w` tensor for layers created with `use_bias=False`.

    Args:
      input_signature: `ShapeDtype` instance characterizing the input this layer
          should compute on.
    """
    shape_w = (input_signature.shape[-1], self._n_units)
    shape_b = (self._n_units,)
    rng_w, rng_b = fastmath.random.split(self.rng, 2)
    w = self._kernel_initializer(shape_w, rng_w)
    if self._use_bfloat16:
      w = w.astype(jnp.bfloat16)

    if self._use_bias:
      b = self._bias_initializer(shape_b, rng_b)
      if self._use_bfloat16:
        b = b.astype(jnp.bfloat16)
      self.weights = (w, b)
    else:
      self.weights = w


# The output tensor has the same shape as the input tensor, but with added
# dimension at the end. This dimension size corresponds to embedding depth.
@assert_shape('...->...d')
class Embedding(base.Layer):
  """Trainable layer that maps discrete tokens/IDs to vectors.

  Embedding layers are commonly used to map discrete data, like words in NLP,
  into vectors. Here is a canonical example::

      vocab_size = 5
      word_ids = np.array([1, 2, 3, 4], dtype=np.int32)  # word_ids < vocab_size
      embedding_layer = tl.Embedding(vocab_size, 32)
      embedding_layer.init(trax.shapes.signature(word_ids))
      embedded = embedding_layer(word_ids)  # embedded.shape = (4, 32)
  """

  def __init__(self,
               vocab_size,
               d_feature,
               use_bfloat16=False,
               kernel_initializer=init.ScaledInitializer(
                   out_dim=-1, in_dim=-2, scale=1., mode='fan_out',
                   distribution='uniform')):
    """Returns an embedding layer with given vocabulary size and vector size.

    The layer clips input values (token IDs) to the range `[0, vocab_size)`.
    That is, negative token IDs all clip to `0` before being mapped to a
    vector, and token IDs with value `vocab_size` or greater all clip to
    `vocab_size - 1` before being mapped to a vector.

    Args:
      vocab_size: Size of the input vocabulary. The layer will assign a unique
        vector to each id in `range(vocab_size)`.
      d_feature: Dimensionality/depth of the output vectors.
      use_bfloat16: If `True`, use bfloat16 weights instead of the default
        float32; this can save memory but may (rarely) lead to numerical issues.
      kernel_initializer: Function that creates (random) initial vectors for
        the embedding.
    """
    # TODO(jonni): is the clipping behavior what we want going forward?
    super().__init__(name=f'Embedding_{vocab_size}_{d_feature}')
    self._d_feature = d_feature  # feature dimensionality
    self._vocab_size = vocab_size
    self._use_bfloat16 = use_bfloat16
    self._kernel_initializer = kernel_initializer

  def forward(self, x):
    """Returns embedding vectors corresponding to input token IDs.

    Args:
      x: Tensor of token IDs.

    Returns:
      Tensor of embedding vectors.
    """
    embedded = jnp.take(self.weights, x, axis=0, mode='clip')
    if self._use_bfloat16:  # Return float32 activations w/ bfloat16 weights.
      embedded = embedded.astype(jnp.float32)
    return embedded

  def init_weights_and_state(self, input_signature):
    """Randomly initializes this layer's weights."""
    del input_signature
    shape_w = (self._vocab_size, self._d_feature)
    # TODO(lukaszkaiser): do we split self.rng for consistency? Add a method?
    w = self._kernel_initializer(shape_w, self.rng)
    if self._use_bfloat16:
      w = w.astype(jnp.bfloat16)
    self.weights = w


@assert_shape('...->...')  # The output and input shapes are the same.
class Dropout(base.Layer):
  """A layer that stochastically ignores a subset of inputs each training step.

  In training, to compensate for the fraction of input values dropped (`rate`),
  all surviving values are multiplied by `1 / (1 - rate)`.

  The parameter `shared_axes` allows to specify a list of axes on which
  the mask will be shared: we will use size 1 on those axes for dropout mask
  and broadcast it. Sharing reduces randomness, but can save memory.

  This layer is active only during training (`mode='train'`). In other
  circumstances it is a no-op.

  Originally introduced in the paper "Dropout: A Simple Way to Prevent Neural
  Networks from Overfitting" available under the following link:
  https://www.cs.toronto.edu/~hinton/absps/JMLRdropout.pdf
  """

  def __init__(self, rate=0.0, shared_axes=None, mode='train'):
    """Creates a dropout layer with the given target drop rate.

    Args:
      rate: Stochastic rate (probability) for dropping an activation value
          from the preceding layer (setting it to zero).
      shared_axes: List of axes on which the mask is shared.
      mode: If `'train'`, this layer will perform dropout; else, it will pass
          all values through unaltered.
    """
    super().__init__()
    self._initial_rate = rate
    self._shared_axes = [] if shared_axes is None else shared_axes
    self._mode = mode

  def init_weights_and_state(self, input_signature):
    """Sets layer-specific internal state."""
    del input_signature
    self.state = jnp.array(self._initial_rate)

  def forward(self, x):
    """Executes this layer as part of a forward pass through the model.

    Args:
      x: Tensor of activations.

    Returns:
      Tensor of same shape and dtype as the input.
    """
    if self._mode != 'train':
      return x
    state, rng = self.state, self.rng
    rate = self._initial_rate
    if isinstance(state, dict) and self._name in state:
      rate = state[self._name]
    if rate == 0.0:
      return x
    mask_shape = list(x.shape)
    for axis in self._shared_axes:
      mask_shape[axis] = 1
    keep_prob = 1.0 - rate
    keep = fastmath.random.bernoulli(rng, keep_prob, tuple(mask_shape))
    mask = keep.astype(x.dtype) / keep_prob
    return x * mask


class Weights(base.Layer):
  """Learnable weights as a layer.

  It takes no input and returns a single tensor: weights.
  """

  def __init__(self, initializer, shape=tuple(), use_bfloat16=False):
    """Returns a learnable tensor of shape `shape`.

    Args:
      initializer: Function taking shape and rng as arguments.
      shape: Shape of the learnable weights.
      use_bfloat16: If `True`, use bfloat16 weights instead of the default
        float32; this can save memory but may (rarely) lead to numerical issues.
    """
    super().__init__(name=f'Weights_{shape}', n_in=0, n_out=1)
    self._shape = shape
    self._initializer = initializer
    self._use_bfloat16 = use_bfloat16

  def forward(self, x):
    """Executes this layer as part of a forward pass through the model.

    Args:
      x: Tensor of same shape and dtype as the input signature used to
          initialize this layer.

    Returns:
      Tensor with previously specified shape and dtype.
    """
    del x  # Unused. There is no input to this layer.
    return self.weights

  def init_weights_and_state(self, input_signature):
    """Returns newly initialized weights for this layer.

    Weights is a single  `w` tensor with previously specified shape.

    Args:
      input_signature: `ShapeDtype` instance characterizing the input this layer
          should compute on. Unused.
    """
    del input_signature  # Unused. There is no input to this layer.
    self.weights = self._initializer(self._shape, self.rng)
    if self._use_bfloat16:
      self.weights = self.weights.astype(jnp.bfloat16)


def PrintShape(n_in=1, msg=''):
  """Prints the shapes of `n_in` inputs and returns then unchanged."""
  def Fwd(xs):
    def format_shape(x):  # pylint: disable = invalid-name
      return str(x.shape) + f'[{x.dtype}]'
    if n_in > 1:
      shapes_and_dtypes = ', '.join([format_shape(x) for x in xs])
    else:
      shapes_and_dtypes = format_shape(xs)
    info = f'PrintShape: {msg}: [{shapes_and_dtypes}]'
    print(info)
    logging.info(info)
    return xs
  return base.PureLayer(Fwd, n_in=n_in, n_out=n_in, name=f'PrintShape_{n_in}')


class SummaryImage(base.Layer):
  """A layer receiving a tensor, and adding it to TensorBoard as an image.

  It takes an input and returns it unchanged. It stores this input as a state to
  be used as a metric in TensorBoard.
  It converts a tensor to a scalar by running a given aggregation function (mean
  by default). On TensorBoard, results for each device will be reported
  separately.
  """

  def __init__(self, name, n_in, num_summaries=5,
               recover_fn=None):
    """Takes a tensor and returns it.

    Args:
      name: Name of the metric to be reported.
      n_in: Number of inputs.
      num_summaries: Number of images to show.
      recover_fn: the function for converting a tensor to a dipslayable image.
    """
    super().__init__(name=f'Summary_{name}', n_in=n_in, n_out=n_in)
    name = 'summary_' + name
    self._name = name
    self._num_summaries = num_summaries
    self._recover_fn = recover_fn

  def forward(self, x):
    """Executes this layer as part of a forward pass through the model.

    Args:
      x: Tensor of same shape and dtype as the input signature used to
          initialize this layer.

    Returns:
      Tensor with previously specified shape and dtype.
    """
    self.state = {}
    batch_size = x[0].shape[0]
    num_images = min(self._num_summaries, batch_size)
    for s in range(num_images):
      images = []
      for i in range(self._n_in):
        images.append(
            self._recover_fn(x[i][s]) if self._recover_fn else x[i][s])
      self.state[self._name + str(s)] = jnp.concatenate(images, axis=0)
    return x[:self._n_in]

  def init_weights_and_state(self, input_signature):
    """Returns newly initialized weights for this layer.

    Weights is a single  `w` tensor with previously specified shape.

    Args:
      input_signature: `ShapeDtype` instance characterizing the input this layer
          should compute on. Unused.
    """
    del input_signature  # Unused.
    self.weights = ()
    self.state = {self._name: jnp.array(0.)}


class SummaryScalar(base.Layer):
  """A layer receiving a tensor, and adding it to TensorBoard as a scalar.

  It takes an input and returns it unchanged. It stores this input as a state to
  be used as a metric in TensorBoard.
  It converts a tensor to a scalar by running a given aggregation function (mean
  by default). On TensorBoard, results for each device will be reported
  separately.
  """

  def __init__(self, name, aggregation_fun=jnp.mean):
    """Takes a tensor and returns it.

    Args:
      name: Name of the metric to be reported.
      aggregation_fun: Aggregation function to be used.
    """
    super().__init__(name=f'Summary_{name}', n_in=1, n_out=1)
    name = 'summary_' + name
    self._name = name
    self._aggregation_fun = aggregation_fun

  def forward(self, x):
    """Executes this layer as part of a forward pass through the model.

    Args:
      x: Tensor of same shape and dtype as the input signature used to
          initialize this layer.

    Returns:
      Tensor with previously specified shape and dtype.
    """
    self.state = {self._name: self._aggregation_fun(x)}
    return x

  def init_weights_and_state(self, input_signature):
    """Returns newly initialized weights for this layer.

    Weights is a single  `w` tensor with previously specified shape.

    Args:
      input_signature: `ShapeDtype` instance characterizing the input this layer
          should compute on. Unused.
    """
    del input_signature  # Unused.
    self.weights = ()
    self.state = {self._name: jnp.array(0.)}


class RandomUniform(base.Layer):
  """Layer returning a tensor with random values distributed uniformly."""

  def __init__(self, min_val=0.0, max_val=1.0, shape=(), dtype=jnp.float32,
               sync=False):
    """Layer returning a tensor with random values distributed uniformly.

    Args:
      min_val: Lower end of uniform distribution.
      max_val: Upper end of uniform distribution.
      shape: Shape of the tensor to return. Values are sampled independently.
      dtype: Type of value to return.
      sync: Whether to synchronise `rng` across devices.
    """
    super().__init__(n_in=0, n_out=1)
    self._min_val = min_val
    self._max_val = max_val
    self._shape = shape
    self._dtype = dtype
    self._sync = sync

  def forward(self, xs):
    """Executes this layer as part of a forward pass through the model.

    Args:
      xs: Unused tensors.

    Returns:
      Random uniform tensor of the shape and type specified in constructor.
    """
    rng = self._get_conditionally_synced_rng()
    result = fastmath.random.uniform(
        rng, self._shape, self._dtype, self._min_val, self._max_val)
    return result

  def _get_conditionally_synced_rng(self):
    if self._sync and fastmath.global_device_count() > 1:
      return fastmath.psum(self.rng, 'batch')
    else:
      return self.rng


class LocallyConnected1d(base.Layer):
  """Locally-connected layer for 1D inputs.

  The LocallyConnected1d layer applies a different set of filters to each patch
  of the input. This is similar to applying a convolution layer, except that
  locally-connected layer uses a different set of weights for each patch.

  The size of patch is determined by the kernel size. The stride is currently
  not modifiable and set to one. This means for the input of shape (..., L, D)
  the output shape for paddings 'SAME' and 'WRAP' will be (..., L, filters) and
  for padding 'VALID' (..., L-kernel_size+1, filters); where L is the number of
  "pixels" or "steps" in the input, D is the size of the embedding.

  Note that, since the weights for different patches are not shared, the number
  of "pixels" or "steps" cannot change after calling init_weights_and_state.
  This is because each "pixel" is assigned its own set of weights.
  """

  def __init__(self, filters, kernel_size,
               kernel_initializer=init.GlorotUniformInitializer(),
               bias_initializer=init.RandomNormalInitializer(1e-6),
               use_bias=True, padding='VALID'):
    """Returns a locally-connected conv-like layer.

    Args:
      filters: Number of output filters in the convolution.
      kernel_size: A length of the convolution window. Must be an odd number.
      kernel_initializer: Function that creates a matrix of (random) initial
          connection weights `W` for the layer.
      bias_initializer: Function that creates a vector of (random) initial
          bias weights `b` for the layer.
      use_bias: If `True`, the layer uses a bias vector.
      padding: The type of padding to use; must be 'VALID', 'SAME', or 'WRAP'.
    """
    super().__init__(name=f'LocallyConnected1d_{filters}_{kernel_size}')
    self._filters = filters
    self._kernel_size = kernel_size
    assert self._kernel_size % 2 == 1  # kernel size has to be odd
    self._kernel_initializer = kernel_initializer
    self._bias_initializer = bias_initializer
    self._use_bias = use_bias
    self._padding = padding

  def forward(self, x):
    """Executes this layer as part of a forward pass through the model.

    Args:
      x: Tensor of same shape and dtype as the input signature used to
          initialize this layer.

    Returns:
      Tensor of same shape and dtype as the input, except the final dimension
      is the layer's `filters` value, and the second to last dimension is
      shrinked if 'VALID' padding is used with kernel_size bigger than one.
    """
    if self._use_bias:
      if not isinstance(self.weights, (tuple, list)):
        raise ValueError(f'Weights should be a (w, b) tuple or list; '
                         f'instead got: {self.weights}')
      w, b = self.weights
    else:
      w = self.weights

    linear_results_before_shifting = jnp.einsum(
        '...lp,lkpd->...lkd', x, w)
    # TODO(jaszczur): this could be run after padding for better efficiency

    if self._kernel_size == 1:
      # With kernel size 1 we don't have to split or shift anything.
      linear_result = jnp.squeeze(linear_results_before_shifting, axis=-2)
    else:
      # We computed a result for every "pixel", but each direction from the
      # receptive field (there are 'self._kernel_size' such directions) must be
      # shifted by a different amount. The easiest way to do it is to split
      # the tensor to 'self._kernel_size' smaller tensors, shift each one
      # appropriately, and then sum them together.
      split_shifting_linear_results = jnp.split(
          linear_results_before_shifting, self._kernel_size, axis=-2)

      for i in range(self._kernel_size):
        # Each tensor has to be shifted a different amount.
        if self._padding == 'WRAP':
          # We can shift by padding and cutting. With 'wrap' padding we
          # essentially have a torus.
          padding = [(0, 0) for i in split_shifting_linear_results[i].shape]
          padding[-3] = ((self._kernel_size - 1) - i, i)
          split_shifting_linear_results[i] = jnp.pad(
              split_shifting_linear_results[i], padding, mode='wrap')
          split_shifting_linear_results[i] = split_shifting_linear_results[i][
              ..., (self._kernel_size-1)//2:-(self._kernel_size-1)//2, :, :]
        elif self._padding == 'SAME':
          # We can shift by padding and cutting.
          padding = [(0, 0) for i in split_shifting_linear_results[i].shape]
          padding[-3] = ((self._kernel_size - 1) - i, i)
          split_shifting_linear_results[i] = jnp.pad(
              split_shifting_linear_results[i], padding)
          split_shifting_linear_results[i] = split_shifting_linear_results[i][
              ..., (self._kernel_size-1)//2:-(self._kernel_size-1)//2, :, :]
          # TODO(jaszczur): improve efficiency by not padding things to cut
        elif self._padding == 'VALID':
          # We don't need to shift - just cut the leftmost and rightmost values.
          cut_left = (self._kernel_size - 1) - i
          cut_right = split_shifting_linear_results[i].shape[-3] - i
          split_shifting_linear_results[i] = split_shifting_linear_results[i][
              ..., cut_left:cut_right, :, :]
        else:
          raise ValueError(f'Invalid padding {self._padding}')
      # After shifting.
      shifted_linear_results = jnp.concatenate(split_shifting_linear_results,
                                               axis=-2)
      linear_result = jnp.sum(shifted_linear_results, axis=-2)

    if self._use_bias:
      return linear_result + b
    else:
      return linear_result

  def init_weights_and_state(self, input_signature):
    """Randomly initializes this layer's weights.

    Weights are a `(w, b)` tuple for layers created with `use_bias=True` (the
    default case), or a `w` tensor for layers created with `use_bias=False`.

    Args:
      input_signature: `ShapeDtype` instance characterizing the input this layer
          should compute on.
    """
    shape_w = (input_signature.shape[-2], self._kernel_size,
               input_signature.shape[-1], self._filters)
    if self._padding == 'VALID':
      shape_b = (input_signature.shape[-2] - self._kernel_size + 1,
                 self._filters,)
    else:
      shape_b = (input_signature.shape[-2], self._filters,)

    rng_w, rng_b = fastmath.random.split(self.rng, 2)
    w = self._kernel_initializer(shape_w, rng_w, nonreceptive_dims=[0])

    if self._use_bias:
      b = self._bias_initializer(shape_b, rng_b)
      self.weights = (w, b)
    else:
      self.weights = w


def Flatten(n_axes_to_keep=1):
  """Returns a layer that combines one or more trailing axes of a tensor.

  Flattening keeps all the values of the input tensor, but reshapes it by
  collapsing one or more trailing axes into a single axis. For example, a
  `Flatten(n_axes_to_keep=2)` layer would map a tensor with shape
  `(2, 3, 5, 7, 11)` to the same values with shape `(2, 3, 385)`.

  Args:
    n_axes_to_keep: Number of leading axes to leave unchanged when reshaping;
        collapse only the axes after these.
  """
  layer_name = f'Flatten_keep{n_axes_to_keep}'
  def f(x):  # pylint: disable=invalid-name
    in_rank = len(x.shape)
    if in_rank <= n_axes_to_keep:
      raise ValueError(f'Input rank ({in_rank}) must exceed the number of '
                       f'axes to keep ({n_axes_to_keep}) after flattening.')
    shape = x.shape
    if isinstance(shape, tf.TensorShape):
      shape = tuple(shape.as_list())
    return jnp.reshape(x, (shape[:n_axes_to_keep] + (-1,)))
  return Fn(layer_name, f)


def LogSoftmax(axis=-1):
  """Returns a layer that applies log softmax along one tensor axis.

  Note that the implementation actually computes x - LogSumExp(x),
  which is mathematically equal to LogSoftmax(x).

  `LogSoftmax` acts on a group of values and normalizes them to look like a set
  of log probability values. (Probability values must be non-negative, and as
  a set must sum to 1. A group of log probability values can be seen as the
  natural logarithm function applied to a set of probability values.)

  Args:
    axis: Axis along which values are grouped for computing log softmax.
  """
  return Fn('LogSoftmax', lambda x: log_softmax(x, axis=axis))


def LogSumExp(axis=-1):
  """Returns a layer that computes log(sum(exp(x))) along one tensor axis.

  Args:
    axis: Axis along which values are grouped for computing log-sum-exp.
  """
  return Fn('LogSumExp',
            lambda x: fastmath.logsumexp(x, axis=axis, keepdims=True))


def Softmax(axis=-1):
  """Returns a layer that applies softmax along one tensor axis.

  `Softmax` acts on a group of values and normalizes them to look like a set
  of probability values. (Probability values must be non-negative, and as a
  set must sum to 1.)

  Args:
    axis: Axis along which values are grouped for computing softmax.
  """
  return Fn('Softmax',
            lambda x: jnp.exp(log_softmax(x, axis=axis)))


def ToFloat():
  """Returns a layer that changes the dtype of a tensor to `float32`."""
  return Fn('ToFloat', lambda x: x.astype(np.float32))


def Mean(axis=-1, keepdims=False):
  """Returns a layer that computes mean values using one tensor axis.

  `Mean` uses one tensor axis to form groups of values and replaces each group
  with the mean value of that group. The resulting values can either remain
  in their own size 1 axis (`keepdims=True`), or that axis can be removed from
  the overall tensor (default `keepdims=False`), lowering the rank of the
  tensor by one.

  Args:
    axis: Axis along which values are grouped for computing a mean.
    keepdims: If `True`, keep the resulting size 1 axis as a separate tensor
        axis; else, remove that axis.
  """
  return Fn('Mean', lambda x: jnp.mean(x, axis=axis, keepdims=keepdims))


def Min(axis=-1, keepdims=False):
  """Returns a layer that applies min along one tensor axis.

  Args:
    axis: Axis along which values are grouped for computing minimum.
    keepdims: If `True`, keep the resulting size 1 axis as a separate tensor
        axis; else, remove that axis.
  """
  return Fn('Min', lambda x: jnp.min(x, axis, keepdims=keepdims))


def Max(axis=-1, keepdims=False):
  """Returns a layer that applies max along one tensor axis.

  Args:
    axis: Axis along which values are grouped for computing maximum.
    keepdims: If `True`, keep the resulting size 1 axis as a separate tensor
        axis; else, remove that axis.
  """
  return Fn('Max', lambda x: jnp.max(x, axis, keepdims=keepdims))


def Sum(axis=None, keepdims=False):
  """Returns a layer that computes sums using one tensor axis.

  `Sum` uses one tensor axis to form groups of values and replaces each group
  with the sum of that group. The resulting sum values can either remain in
  their own size 1 axis (`keepdims=True`), or that axis can be removed from the
  overall tensor (default `keepdims=False`), lowering the rank of the tensor by
  one.

  Args:
    axis: Axis along which values are grouped for computing a sum; if None,
        compute sum over all elements in tensor.
    keepdims: If `True`, keep the resulting size 1 axis as a separate tensor
        axis; else, remove that axis.
  """
  return Fn('Sum', lambda x: jnp.sum(x, axis=axis, keepdims=keepdims))


def ThresholdToBinary(threshold=.5):
  """Returns a layer that thresholds inputs to yield outputs in {0, 1}."""
  def f(model_output):  # pylint: disable=invalid-name
    return (model_output > threshold).astype(jnp.int32)
  return Fn('ThresholdToBinary', f)


def ArgMax(axis=-1):
  """Returns a layer that calculates argmax along the given axis."""
  def f(model_output):  # pylint: disable=invalid-name
    return jnp.argmax(model_output, axis=axis)
  return Fn('ArgMax', f)


@assert_shape('...->...')  # The output and input shapes are the same.
def Negate():
  """Returns a layer that computes the element-wise negation of a tensor."""
  return Fn('Negate', lambda x: -x)


@assert_shape('...->...')  # The output and input shapes are the same.
def StopGradient():
  """Returns an identity layer with a stop gradient."""
  return Fn('StopGradient', lambda x: fastmath.stop_gradient(x))  # pylint: disable=unnecessary-lambda


def one_hot(x, n_categories, dtype=jnp.float32):  # pylint: disable=invalid-name
  """Makes a one-hot array (n+1 dims) from an int-categorical array (n dims)."""
  indices_less_than_n = jnp.arange(n_categories)
  return jnp.array(x[..., jnp.newaxis] == indices_less_than_n, dtype)


def log_softmax(x, axis=-1):  # pylint: disable=invalid-name
  """Transforms activation vectors to log-probability vectors.

  Log probability vectors are derived by, in effect, applying softmax to raw
  activation vectors and then applying log element-wise. The actual
  implementation uses a mathematically valid simplification of this.

  Args:
    x: An ndarray with activation vectors along the given axis.
    axis: Axis along which values are grouped for computing log softmax.

  Returns:
    An ndarray containing log-probability vectors derived from the raw
    activation vectors in `x`.
  """
  return x - fastmath.logsumexp(x, axis=axis, keepdims=True)


def log_gaussian_pdf(x, mu, sigma):  # pylint: disable=invalid-name
  """Returns `log N(x | mu, sigma)`.

  Args:
    x: <tbd>
    mu: <tbd>
    sigma: <tbd>
  """
  a = mu.shape[-1] * jnp.log(2 * jnp.pi)
  _, b = jnp.linalg.slogdet(sigma)
  y = jnp.linalg.solve(sigma, x - mu)
  y = jnp.expand_dims(y, axis=-1)
  xm = jnp.expand_dims(x - mu, axis=-2)
  c = jnp.matmul(xm, y)
  c = jnp.squeeze(jnp.squeeze(c, axis=-1), axis=-1)
  return -0.5 * (a + b + c)


def log_gaussian_diag_pdf(x, mu, diag_sigma):  # pylint: disable=invalid-name
  """Returns `log N(x | mu, eye(diag_sigma))`.

  Args:
    x: <tbd>
    mu: <tbd>
    diag_sigma: <tbd>
  """
  a = mu.shape[-1] * jnp.log(2 * jnp.pi)
  b = jnp.sum(jnp.log(diag_sigma), axis=-1)
  y = x - mu / diag_sigma
  y = jnp.expand_dims(y, axis=-1)
  xm = jnp.expand_dims(x - mu, axis=-2)
  c = jnp.matmul(xm, y)
  c = jnp.squeeze(jnp.squeeze(c, axis=-1), axis=-1)
  return -0.5 * (a + b + c)


def multigaussian_loss(preds, targets, ngauss=1):  # pylint: disable=invalid-name
  """Returns a mixture of gaussians loss.

  Args:
    preds: <tbd>
    targets: <tbd>
    ngauss: <tbd>
  """
  ndims = targets.shape[-1]
  logits = preds[:, :ngauss]
  mus = preds[:, ngauss:ngauss*(ndims + 1)]
  sigmas = preds[:, ngauss(ndims + 1):]
  sigmas = sigmas * sigmas + 1e-6  # Make positive.
  loglogits = logits - fastmath.logsumexp(logits, axis=-1, keepdims=True)
  mus = jnp.reshape(mus, [-1, ngauss, ndims])
  sigmas = jnp.reshape(sigmas, [-1, ngauss, ndims])
  targets = jnp.reshape(targets, [-1, 1, ndims])
  glogprobs = log_gaussian_diag_pdf(targets, mus, sigmas)
  return fastmath.logsumexp(loglogits + glogprobs, axis=-1)


# TODO(jonni): Rename to log_softmax_sample.
def logsoftmax_sample(log_probs, temperature=1.0):  # pylint: disable=invalid-name
  """Returns a sample from a log-softmax output, with temperature.

  Args:
    log_probs: Logarithms of probabilities (often coming from LogSoftmax)
    temperature: For scaling before sampling (1.0 = default, 0.0 = pick argmax)
  """
  # This is equivalent to sampling from a softmax with temperature.
  u = np.random.uniform(low=1e-6, high=1.0 - 1e-6, size=log_probs.shape)
  g = -np.log(-np.log(u))
  return np.argmax(log_probs + g * temperature, axis=-1)



# coding=utf-8
# Copyright 2021 The Trax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Classes for RL training in Trax."""

import functools

from trax import data
from trax import layers as tl
from trax import supervised
from trax.fastmath import numpy as jnp
from trax.fastmath import stop_gradient
from trax.rl import actor_critic
from trax.rl import distributions
from trax.rl import rl_layers
from trax.rl import training as rl_training
from trax.supervised import lr_schedules as lr


# pylint: disable=g-long-lambda
class ActorCriticJointAgent(rl_training.Agent):
  """Trains a joint policy-and-value model using actor-critic methods."""

  def __init__(self, task,
               joint_model=None,
               optimizer=None,
               lr_schedule=lr.multifactor,
               batch_size=64,
               train_steps_per_epoch=500,
               supervised_evals_per_epoch=1,
               supervised_eval_steps=1,
               n_trajectories_per_epoch=50,
               max_slice_length=1,
               normalize_advantages=True,
               output_dir=None,
               n_replay_epochs=1):
    """Configures the joint trainer.

    Args:
      task: RLTask instance, which defines the environment to train on.
      joint_model: Trax layer, representing the joint policy and value model.
      optimizer: the optimizer to use to train the joint model.
      lr_schedule: learning rate schedule to use to train the joint model/.
      batch_size: batch size used to train the joint model.
      train_steps_per_epoch: how long to train the joint model in each RL epoch.
      supervised_evals_per_epoch: number of value trainer evaluations per RL
          epoch - only affects metric reporting.
      supervised_eval_steps: number of value trainer steps per evaluation -
          only affects metric reporting.
      n_trajectories_per_epoch: how many trajectories to collect per epoch.
      max_slice_length: the maximum length of trajectory slices to use.
      normalize_advantages: if True, then normalize advantages - currently
          implemented only in PPO.
      output_dir: Path telling where to save outputs (evals and checkpoints).
      n_replay_epochs: how many last epochs to take into the replay buffer;
           > 1 only makes sense for off-policy algorithms.
    """
    super().__init__(
        task,
        n_trajectories_per_epoch=n_trajectories_per_epoch,
        output_dir=output_dir,
    )
    self._batch_size = batch_size
    self._train_steps_per_epoch = train_steps_per_epoch
    self._supervised_evals_per_epoch = supervised_evals_per_epoch
    self._supervised_eval_steps = supervised_eval_steps
    self._n_trajectories_per_epoch = n_trajectories_per_epoch
    self._max_slice_length = max_slice_length
    self._policy_dist = distributions.create_distribution(task.action_space)
    self._lr_schedule = lr_schedule()
    self._optimizer = optimizer
    self._normalize_advantages = normalize_advantages
    self._n_replay_epochs = n_replay_epochs
    self._task.set_n_replay_epochs(n_replay_epochs)

    # Inputs to the joint model are produced by self.batches_stream.
    self._inputs = data.inputs.Inputs(
        train_stream=lambda _: self.batches_stream())

    self._joint_model = functools.partial(
        joint_model,
        policy_distribution=self._policy_dist,
    )

    # This is the joint Trainer that will be used to train the policy model.
    # * inputs to the trainer come from self.batches_stream
    # * outputs are passed to self._joint_loss
    self._trainer = supervised.Trainer(
        model=self._joint_model,
        optimizer=self._optimizer,
        lr_schedule=self._lr_schedule,
        loss_fn=self.joint_loss,
        inputs=self._inputs,
        output_dir=output_dir,
        metrics={'joint_loss': self.joint_loss,
                 'advantage_mean': self.advantage_mean,
                 'advantage_norm': self.advantage_norm,
                 'value_loss': self.value_loss,
                 'explained_variance': self.explained_variance,
                 'log_probs_mean': self.log_probs_mean,
                 'preferred_move': self.preferred_move})
    self._eval_model = tl.Accelerate(
        self._joint_model(mode='eval'), n_devices=1)
    example_batch = next(self.batches_stream())
    self._eval_model.init(example_batch)

  def close(self):
    self._trainer.close()
    super().close()

  def batches_stream(self):
    """Use self.task to create inputs to the policy model."""
    return NotImplementedError

  @property
  def joint_loss(self):
    """Joint policy and value loss layer."""
    return NotImplementedError

  @property
  def advantage_mean(self):
    """Mean of advantages."""
    def f(dist_inputs, values, returns):
      del dist_inputs
      return jnp.mean(returns - values)
    return tl.Fn('AdvantageMean', f)

  @property
  def advantage_norm(self):
    """Norm of advantages."""
    def f(dist_inputs, values, returns):
      del dist_inputs
      return jnp.linalg.norm(returns - values)
    return tl.Fn('AdvantageNorm', f)

  @property
  def value_loss(self):
    """Value loss - so far generic for all A2C."""
    def f(dist_inputs, values, returns):
      del dist_inputs
      return rl_layers.ValueLoss(values, returns, self._value_loss_coeff)
    return tl.Fn('ValueLoss', f)

  @property
  def explained_variance(self):
    """Explained variance metric."""
    def f(dist_inputs, values, returns):
      del dist_inputs
      return rl_layers.ExplainedVariance(values, returns)
    return tl.Fn('ExplainedVariance', f)

  @property
  def log_probs_mean(self):
    """Mean of log_probs aka dist_inputs."""
    def f(dist_inputs, values):
      del values
      return jnp.mean(dist_inputs)
    return tl.Fn('LogProbsMean', f)

  @property
  def preferred_move(self):
    """Preferred move - the mean of selected moves."""
    def f(dist_inputs, values):
      del values
      return rl_layers.PreferredMove(dist_inputs, self._policy_dist.sample)
    return tl.Fn('PreferredMove', f)

  def policy(self, trajectory, temperature=1.0):
    """Chooses an action to play after a trajectory."""
    model = self._eval_model
    model.replicate_weights(self._trainer.model_weights)
    # The two lines below along with the copying
    # before return make the TPU happy
    tr_slice = trajectory.suffix(self._max_slice_length)
    trajectory_np = tr_slice.to_np(timestep_to_np=self.task.timestep_to_np)
    # Add batch dimension to trajectory_np and run the model.
    pred = model(trajectory_np.observation[None, ...])[0]
    # Pick element 0 from the batch (the only one), last (current) timestep.
    pred = pred[0, -1, :]
    sample = self._policy_dist.sample(pred, temperature=temperature)
    return (sample.copy(), pred.copy())

  def train_epoch(self):
    """Trains RL for one epoch."""
    n_evals = rl_training.remaining_evals(
        self._trainer.step,
        self._epoch,
        self._train_steps_per_epoch,
        self._supervised_evals_per_epoch)
    for _ in range(n_evals):
      self._trainer.train_epoch(
          self._train_steps_per_epoch // self._supervised_evals_per_epoch,
          self._supervised_eval_steps)


class PPOJoint(ActorCriticJointAgent):
  """The Proximal Policy Optimization Algorithm aka PPO.

  Trains policy and value models using the PPO algortithm.
  """

  # TODO(henrykm): make on_policy more generic
  # (currently epochs are passed manually)
  on_policy = True

  def __init__(self, task, epsilon=0.2, value_loss_coeff=0.1,
               entropy_coeff=0.01, **kwargs):
    """Configures the PPO Trainer."""
    self._epsilon = epsilon
    self._value_loss_coeff = value_loss_coeff
    self._entropy_coeff = entropy_coeff
    super().__init__(task, **kwargs)
    self._trainer = supervised.Trainer(
        model=self._joint_model,
        optimizer=self._optimizer,
        lr_schedule=self._lr_schedule,
        loss_fn=self.joint_loss,
        inputs=self._inputs,
        output_dir=self._output_dir,
        metrics={'joint_loss': self.joint_loss,
                 'advantage_mean': self.advantage_mean,
                 'advantage_norm': self.advantage_norm,
                 'value_loss': self.value_loss,
                 'explained_variance': self.explained_variance,
                 'log_probs_mean': self.log_probs_mean,
                 'entropy_loss': self.entropy_loss,
                 'probs_ratio_mean': self.probs_ratio_mean,
                 'unclipped_objective_mean': self.unclipped_objective_mean,
                 'clipped_objective_mean': self.clipped_objective_mean,
                 'ppo_objective_mean': self.ppo_objective_mean,
                 'clip_fraction': self.clip_fraction,
                 'preferred_move': self.preferred_move,
                 'approximate_kl_divergence': self.approximate_kl_divergence})

  def batches_stream(self):
    """Use the RLTask self._task to create inputs to the value model."""
    for np_trajectory in self._task.trajectory_batch_stream(
        self._batch_size, max_slice_length=self._max_slice_length, epochs=[-1]):
      if np_trajectory.dist_inputs is not None:
        old_dist_inputs = np_trajectory.dist_inputs
      else:
        old_dist_inputs = jnp.zeros(
            np_trajectory.reward.shape + (self._policy_dist.n_inputs,)
        )
      old_log_probs = self._policy_dist.log_prob(
          old_dist_inputs, np_trajectory.action
      )
      # Insert an extra depth dimension, so the target shape is consistent with
      # the network output shape.
      yield (np_trajectory.observation,         # Inputs to the value model.
             np_trajectory.return_[:, :, None],
             np_trajectory.done[:, :, None],
             np_trajectory.reward[:, :, None],
             np_trajectory.action,
             old_log_probs,
             np_trajectory.mask)

  @property
  def joint_loss(self):
    """Joint policy and value loss."""
    def f(dist_inputs, values, returns, dones, rewards,
          actions, old_log_probs, mask):
      """Definition of the Proximal Policy Optimization loss."""
      del mask  # TODO(lukaszkaiser): make PPO work with Transformer
      # We have dist_inputs of the shape float32[128,1,18]
      assert len(dist_inputs.shape) == 3, (
          f'dist_inputs.shape was {dist_inputs.shape}'
          f'but expected length of the tensor shape is 3')
      # values of the shape float32[128,1,1]
      # returns of the shape float32[128,1,1]
      # dones of the shape int32[128,1,1]
      # rewards of the shape float32[128,1,1]
      # and old_log_probs of the shape float32[128,1]
      assert values.shape == returns.shape, (
          f'values.shape was {values.shape}'
          f'returns.shape was {returns.shape}')
      assert values.shape == dones.shape, (
          f'values.shape was {values.shape}'
          f'returns.shape was {dones.shape}')
      assert rewards.shape == dones.shape, (
          f'values.shape was {values.shape}'
          f'returns.shape was {dones.shape}')
      assert returns.shape[0:2] == old_log_probs.shape, (
          f'returns.shape was {returns.shape}'
          f'old_log_probs.shape was {old_log_probs.shape}')

      # actions is a tensor of the shape int32[128,1] in the case
      # of discrete actions and float32[128,1,6] in the case of
      # half-cheetah and other continuous actions
      # actions agree with returns/values on the first two coordinates
      # meaning batch and time
      assert actions.shape[0:2] == returns.shape[0:2], (
          f'actions.shape was {actions.shape} and '
          f'returns.shape was {returns.shape}')

      ppo_objective = rl_layers.PPOObjective(
          dist_inputs, stop_gradient(values), returns, dones, rewards,
          actions, old_log_probs,
          log_prob_fun=self._policy_dist.log_prob,
          epsilon=self._epsilon,
          normalize_advantages=self._normalize_advantages)

      # we insist that ppo_objective is a vector of shape [128,1]
      assert len(ppo_objective.shape) == 2, (
          f'ppo_objective was {ppo_objective}')
      # which agrees with returns/values/actions on the first two coordinates
      assert ppo_objective.shape[0:2] == values.shape[0:2], (
          f'ppo_objective.shape was {ppo_objective.shape} and '
          f'values.shape was {values.shape}')

      entropy_loss = rl_layers.EntropyLoss(
          dist_inputs,
          distribution=self._policy_dist,
          coeff=self._entropy_coeff,
      )

      assert jnp.ndim(entropy_loss) == 0, f'entropy_loss was {entropy_loss}'

      l2_value_loss = rl_layers.ValueLoss(
          values, returns, value_loss_coeff=self._value_loss_coeff)

      assert jnp.ndim(l2_value_loss) == 0, f'l2_value_loss was {l2_value_loss}'

      return -ppo_objective.mean() + l2_value_loss - entropy_loss

    return tl.Fn('PPOJointLoss', f)

  # pylint: disable=invalid-name
  @property
  def probs_ratio_mean(self):
    """Joint policy and value loss layer."""
    def ProbsRatioMean(dist_inputs, actions, old_log_probs):
      """Probability Ratio Mean from the PPO algorithm."""
      probs_ratio = rl_layers.ProbsRatio(
          dist_inputs, actions, old_log_probs,
          log_prob_fun=self._policy_dist.log_prob)
      return jnp.mean(probs_ratio)

    def f(dist_inputs, values, returns, dones, rewards, actions, old_log_probs):
      del values, returns, dones, rewards
      return ProbsRatioMean(dist_inputs, actions, old_log_probs)
    return tl.Fn('ProbsRatioMean', f)

  @property
  def clip_fraction(self):
    """Joint policy and value loss layer."""
    def ClipFraction(dist_inputs, actions, old_log_probs):
      """Probability Ratio Mean from the PPO algorithm."""
      probs_ratio = rl_layers.ProbsRatio(
          dist_inputs, actions, old_log_probs,
          log_prob_fun=self._policy_dist.log_prob)
      return jnp.mean(jnp.abs(probs_ratio - 1) > self._epsilon)

    def f(dist_inputs, values, returns, dones, rewards, actions, old_log_probs):
      del values, returns, dones, rewards
      return ClipFraction(dist_inputs, actions, old_log_probs)
    return tl.Fn('ClipFraction', f)
  # pylint: enable=invalid-name

  @property
  def entropy_loss(self):
    """Entropy layer."""
    def f(dist_inputs, values, returns, dones, rewards, actions):
      del values, returns, dones, rewards, actions
      return rl_layers.EntropyLoss(
          dist_inputs,
          distribution=self._policy_dist,
          coeff=self._entropy_coeff,
      )
    return tl.Fn('EntropyLoss', f)

  @property
  def approximate_kl_divergence(self):
    """Approximate KL divergence."""
    def f(dist_inputs, values, returns, dones, rewards,
          actions, old_log_probs):
      del values, returns, dones, rewards
      return rl_layers.ApproximateKLDivergence(
          dist_inputs,
          actions,
          old_log_probs,
          log_prob_fun=self._policy_dist.log_prob)
    return tl.Fn('ApproximateKLDivergence', f)

  @property
  def unclipped_objective_mean(self):
    def f(dist_inputs, values, returns, dones, rewards, actions, old_log_probs):
      """Unclipped objective Mean from the PPO algorithm."""
      del dones, rewards
      advantages = returns - values
      probs_ratio = rl_layers.ProbsRatio(
          dist_inputs, actions, old_log_probs,
          log_prob_fun=self._policy_dist.log_prob)
      # advantages are of the shape [128,1,1]
      # and probs_ratio are of the shape [128,1]
      advantages = advantages.squeeze(axis=2)
      unclipped_objective = rl_layers.UnclippedObjective(
          probs_ratio, advantages)
      return jnp.mean(unclipped_objective)

    return tl.Fn('UnclippedObjectiveMean', f)

  @property
  def clipped_objective_mean(self):
    def f(dist_inputs, values, returns, dones, rewards, actions, old_log_probs):
      """Clipped objective from the PPO algorithm."""
      del dones, rewards
      advantages = returns - values
      probs_ratio = rl_layers.ProbsRatio(
          dist_inputs, actions, old_log_probs,
          log_prob_fun=self._policy_dist.log_prob)
      # advantages are of the shape [128,1,1]
      # and probs_ratio are of the shape [128,1]
      advantages = advantages.squeeze(axis=2)
      clipped_objective = rl_layers.ClippedObjective(
          probs_ratio, advantages, epsilon=self._epsilon)
      return jnp.mean(clipped_objective)

    return tl.Fn('ClippedObjectiveMean', f)

  @property
  def ppo_objective(self):
    """PPO objective with local parameters."""
    def f(dist_inputs, values, returns, dones, rewards, actions, old_log_probs):
      return rl_layers.PPOObjective(
          dist_inputs, values, returns, dones, rewards, actions, old_log_probs,
          log_prob_fun=self._policy_dist.log_prob,
          epsilon=self._epsilon,
          normalize_advantages=self._normalize_advantages)
    return tl.Fn('PPOObjective', f)

  @property
  def ppo_objective_mean(self):
    """PPO objective mean."""
    def f(dist_inputs, values, returns, dones, rewards, actions, old_log_probs):
      """Clipped objective from the PPO algorithm."""
      ppo_objective = rl_layers.PPOObjective(
          dist_inputs, values, returns, dones, rewards, actions, old_log_probs,
          log_prob_fun=self._policy_dist.log_prob,
          epsilon=self._epsilon,
          normalize_advantages=self._normalize_advantages)
      return jnp.mean(ppo_objective)
    return tl.Fn('PPOObjectiveMean', f)


class A2CJoint(ActorCriticJointAgent):
  """The A2C algorithm.

  Trains policy and value models using the A2C algortithm.
  """

  on_policy = True

  def __init__(self, task, value_loss_coeff=0.1,
               entropy_coeff=0.01, **kwargs):
    """Configures the A2C Trainer."""
    self._value_loss_coeff = value_loss_coeff
    self._entropy_coeff = entropy_coeff
    super().__init__(task, **kwargs)
    self._trainer = supervised.Trainer(
        model=self._joint_model,
        optimizer=self._optimizer,
        lr_schedule=self._lr_schedule,
        loss_fn=self.joint_loss,
        inputs=self._inputs,
        output_dir=self._output_dir,
        metrics={'joint_loss': self.joint_loss,
                 'advantage_mean': self.advantage_mean,
                 'advantage_norm': self.advantage_norm,
                 'value_loss': self.value_loss,
                 'explained_variance': self.explained_variance,
                 'log_probs_mean': self.log_probs_mean,
                 'entropy_loss': self.entropy_loss,
                 'a2c_objective_mean': self.a2c_objective_mean,
                 'approximate_kl_divergence': self.approximate_kl_divergence,
                 'preferred_move': self.preferred_move})

  def batches_stream(self):
    """Use the RLTask self._task to create inputs to the value model."""
    for np_trajectory in self._task.trajectory_batch_stream(
        self._batch_size, max_slice_length=self._max_slice_length, epochs=[-1]):
      # Insert an extra depth dimension, so the target shape is consistent with
      # the network output shape.
      yield (np_trajectory.observation,         # Inputs to the value model.
             np_trajectory.return_[:, :, None],
             np_trajectory.done[:, :, None],
             np_trajectory.reward[:, :, None],
             np_trajectory.action,
             jnp.zeros_like(np_trajectory.mask),
             np_trajectory.mask)

  @property
  def joint_loss(self):
    """Joint policy and value loss."""
    def f(dist_inputs, values, returns, dones, rewards,
          actions, old_log_probs, mask):
      """Definition of the A2C loss."""
      del old_log_probs

      # Typically we have dist_inputs of the shape float32[128,1,18]
      assert len(dist_inputs.shape) == 3, (
          f'dist_inputs.shape was {dist_inputs.shape} '
          f'but expected length of the tensor shape is 3')
      # values of the shape float32[128,1,1]
      # returns of the shape float32[128,1,1]
      assert values.shape == returns.shape, (
          f'values.shape was {values.shape}'
          f'returns.shape was (returns.shape)')
      # actions of the shape int32[128,1] in the case of discrete actions
      # and float32[128,1,6] in the case of of half-cheetah
      # actions agree with returns/values on the first two coordinates
      assert actions.shape[0:2] == returns.shape[0:2], (
          f'actions.shape was {actions.shape}'
          f'returns.shape was (returns.shape)')
      # and mask of the shape float32[128,1]
      assert len(mask.shape) == 2, f'mask.shape was {mask.shape}'
      # which agrees with returns/values/actions on the first two coordinates
      assert mask.shape[0:2] == returns.shape[0:2], (
          f'mask.shape was {mask.shape}'
          f'returns.shape was (returns.shape)')

      a2c_objective = rl_layers.A2CObjective(
          dist_inputs,
          stop_gradient(values),
          returns, dones, rewards, actions, mask,
          log_prob_fun=self._policy_dist.log_prob,
          normalize_advantages=self._normalize_advantages)

      # we insist that a2c_objective is a scalar
      assert jnp.ndim(a2c_objective) == 0, f'a2c_objective was {a2c_objective}'

      entropy_loss = rl_layers.EntropyLoss(
          dist_inputs,
          distribution=self._policy_dist,
          coeff=self._entropy_coeff,
      )

      assert jnp.ndim(entropy_loss) == 0, f'entropy_loss was {entropy_loss}'

      l2_value_loss = rl_layers.ValueLoss(
          values, returns, value_loss_coeff=self._value_loss_coeff)

      assert jnp.ndim(l2_value_loss) == 0, f'l2_value_loss was {l2_value_loss}'

      combined_loss = a2c_objective + l2_value_loss - entropy_loss

      return combined_loss

    return tl.Fn('A2CJointLoss', f, n_out=1)

  @property
  def entropy_loss(self):
    """Entropy layer."""
    def f(dist_inputs, values, returns, dones, rewards, actions):
      del values, returns, dones, rewards, actions
      return rl_layers.EntropyLoss(
          dist_inputs,
          distribution=self._policy_dist,
          coeff=self._entropy_coeff,
      )
    return tl.Fn('EntropyLoss', f)

  @property
  def approximate_kl_divergence(self):
    """Approximate KL divergence."""
    def f(dist_inputs, values, returns, dones, rewards,
          actions, old_log_probs):
      del values, returns, dones, rewards
      return rl_layers.ApproximateKLDivergence(
          dist_inputs,
          actions,
          old_log_probs,
          log_prob_fun=self._policy_dist.log_prob)
    return tl.Fn('ApproximateKLDivergence', f)

  @property
  def a2c_objective(self):
    """A2C objective with local parameters."""
    return tl.Fn(
        'A2CObjective',
        lambda dist_inputs, values, returns, dones, rewards, actions, \
        old_log_probs, mask: rl_layers.A2CObjective(
            dist_inputs,
            values,
            returns,
            dones,
            rewards,
            actions,
            mask,
            log_prob_fun=self._policy_dist.log_prob,
            normalize_advantages=self._normalize_advantages),
        n_out=1)

  @property
  def a2c_objective_mean(self):
    """A2C objective mean."""
    def f(dist_inputs, values, returns, dones, rewards,
          actions, old_log_probs, mask):
      """A2C objective mean."""
      # TODO(henrykm): include dones, rewards
      del old_log_probs
      a2c_objective = rl_layers.A2CObjective(
          dist_inputs, values, returns, dones, rewards, actions, mask,
          log_prob_fun=self._policy_dist.log_prob,
          normalize_advantages=self._normalize_advantages)
      return jnp.mean(a2c_objective)
    return tl.Fn('A2CObjectiveMean', f, n_out=1)


class AWRJoint(ActorCriticJointAgent):
  """Trains a joint policy-and-value model using AWR."""

  # TODO(henrykm): value_loss_coeff looks like a common parameter
  def __init__(self, task, value_loss_coeff=0.1, beta=1.0, w_max=20.0,
               thresholds=None, **kwargs):
    """Configures the joint AWR Trainer."""
    self._beta = beta
    self._w_max = w_max
    self._thresholds = thresholds
    self._value_loss_coeff = value_loss_coeff
    super().__init__(task, **kwargs)

  def batches_stream(self):
    """Use the RLTask self._task to create inputs to the value model."""
    for np_trajectory in self._task.trajectory_batch_stream(
        self._batch_size, max_slice_length=self._max_slice_length):
      # Insert an extra depth dimension, so the target shape is consistent with
      # the network output shape.
      yield (np_trajectory.observation,         # Inputs to the value model.
             np_trajectory.return_[:, :, None],  # Targets: regress to returns.
             np_trajectory.action,              # Policy targets: actions.
             np_trajectory.mask)                 # Padding mask.

  @property
  def joint_loss(self):
    """Joint policy and value loss."""

    def f(preds, values, returns, actions, mask):
      advantages = jnp.squeeze(returns - stop_gradient(values), axis=-1)
      logps = self._policy_dist.log_prob(preds, actions)
      awr_loss = actor_critic.AWRLoss(
          beta=self._beta, w_max=self._w_max, thresholds=self._thresholds)(
              (logps, advantages, jnp.zeros_like(logps), mask))
      l2_value_loss = jnp.mean((returns - values)**2) * self._value_loss_coeff
      return awr_loss + l2_value_loss
    return tl.Fn('AWRJointLoss', f)
